/*
 * sim-fast.c - sample fast functional simulator implementation
 *
 * This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin as a part of the Multiscalar Research Project.
 *  
 * The tool suite is currently maintained by Doug Burger and Todd M. Austin.
 * 
 * Copyright (C) 1994, 1995, 1996, 1997 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 *
 * $Id: sim-fast.c,v 1.4 1997/03/11 01:28:36 taustin Exp taustin $
 *
 * $Log: sim-fast.c,v $
 * Revision 1.4  1997/03/11  01:28:36  taustin
 * updated copyright
 * simulation now fails when DLite! support requested
 * long/int tweaks made for ALPHA target support
 * supported added for non-GNU C compilers
 *
 * Revision 1.3  1997/01/06  16:04:21  taustin
 * comments updated
 * SPARC-specific compilation supported deleted (most opts now generalized)
 * new stats and options package support added
 * NO_INSN_COUNT added for conditional instruction count support
 * USE_JUMP_TABLE added to support simple and complex main loops
 *
 * Revision 1.1  1996/12/05  18:52:32  taustin
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
 * This file implements a very fast functional simulator.  This functional
 * simulator implementation is much more difficult to digest than the simpler,
 * cleaner sim-safe functional simulator.  By default, this simulator performs
 * no instruction error checking, as a result, any instruction errors will
 * manifest as simulator execution errors, possibly causing sim-fast to
 * execute incorrectly or dump core.  Such is the price we pay for speed!!!!
 *
 * The following options configure the bag of tricks used to make sim-fast
 * live up to its name.  For most machines, defining all the options results
 * in the fastest functional simulator.
 */

/* no instruction checks, for speed */
#define NO_ICHECKS

/* don't count instructions flag, enabled by default, disable for inst count */
#define NO_INSN_COUNT

/* use a jump table for speed, this is a GNU GCC specific optimization,
   CAVEAT: some versions of GNU GCC core dump when optimizing the jump table
   code with optimization levels higher than -O1 */
/* #define USE_JUMP_TABLE */

#include "misc.h"
#include "ss.h"
#include "regs.h"

/* the following is some macro magic needed to register allocate the
   pointer to the level one memory page map table */

/* the level 1 page table map */
#define mem_table local_mem_table
#define HIDE_MEM_TABLE_DEF
#include "memory.h"
#undef HIDE_MEM_TABLE_DEF
#undef mem_table
extern char *mem_table[MEM_TABLE_SIZE];
#define mem_table local_mem_table

#include "loader.h"
#include "syscall.h"
#include "dlite.h"
#include "sim.h"

#ifndef NO_INSN_COUNT
/* register allocate instruction counter */
unsigned int sim_num_insn = 0;
#else /* NO_INSN_COUNT */
/* no insn or ref count in sim-fast, use sim-safe for instruction counts */
#endif /* !NO_INSN_COUNT */

/* register simulator-specific options */
void
sim_reg_options(struct opt_odb_t *odb)
{
  opt_reg_header(odb, 
"sim-fast: This simulator implements a very fast functional simulator.  This\n"
"functional simulator implementation is much more difficult to digest than\n"
"the simpler, cleaner sim-safe functional simulator.  By default, this\n"
"simulator performs no instruction error checking, as a result, any\n"
"instruction errors will manifest as simulator execution errors, possibly\n"
"causing sim-fast to execute incorrectly or dump core.  Such is the\n"
"price we pay for speed!!!!\n"
		 );
}

/* check simulator-specific option values */
void
sim_check_options(struct opt_odb_t *odb, int argc, char **argv)
{
  if (dlite_active)
    fatal("sim-fast does not support DLite debugging");
}

/* register simulator-specific statistics */
void
sim_reg_stats(struct stat_sdb_t *sdb)
{
#ifndef NO_INSN_COUNT
  stat_reg_uint(sdb, "sim_num_insn",
		"total number of instructions executed",
		&sim_num_insn, 0, NULL);
#endif /* !NO_INSN_COUNT */
  stat_reg_int(sdb, "sim_elapsed_time",
	       "total simulation time in seconds",
	       (int *)&sim_elapsed_time, 0, NULL);
#ifndef NO_INSN_COUNT
  stat_reg_formula(sdb, "sim_inst_rate",
		   "simulation speed (in insts/sec)",
		   "sim_num_insn / sim_elapsed_time", NULL);
#endif /* !NO_INSN_COUNT */
}

/* these are global equivalent defs to the register defs in sim_main(), and
   they are used by the helper functions defined in the `.def' file and
   sim_init() memory accessors because they cannot access the registerized
   version of these vars in sim_main() */
SS_WORD_TYPE *local_regs_R = regs_R;
#undef mem_table
char **local_mem_table = mem_table;
#define mem_table local_mem_table

/* initialize the simulator */
void
sim_init(void)
{
  SS_INST_TYPE inst;

  /* decode all instructions */
  {
    SS_ADDR_TYPE addr;

    if (OP_MAX > 255)
      fatal("cannot do fast decoding, too many opcodes");

    debug("sim: decoding text segment...");
    for (addr=ld_text_base;
	 addr < (ld_text_base+ld_text_size);
	 addr += SS_INST_SIZE)
      {
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, addr);
	inst.a = (inst.a & ~0xff) | (unsigned int)SS_OP_ENUM(SS_OPCODE(inst));
	__UNCHK_MEM_ACCESS(SS_INST_TYPE, addr) = inst;
      }
  }
}

/* print simulator-specific configuration information */
void
sim_aux_config(FILE *stream)
{
  /* nothing currently */
}

/* dump simulator-specific auxiliary simulator statistics */
void
sim_aux_stats(FILE *stream)
{
  /* nada */
}

/* un-initialize simulator-specific state */
void
sim_uninit(void)
{
  /* nada */
}

/*
 * configure the execution engine
 */

/* next program counter */
#define SET_NPC(EXPR)		(local_next_PC = (EXPR))

/* current program counter */
#define CPC			(local_regs_PC)

/* general purpose registers */
#define GPR(N)			(local_regs_R[N])
#define SET_GPR(N,EXPR)		(local_regs_R[N] = (EXPR))

/* floating point registers, L->word, F->single-prec, D->double-prec */
#define FPR_L(N)		(regs_F.l[(N)])
#define SET_FPR_L(N,EXPR)	(regs_F.l[(N)] = (EXPR))
#define FPR_F(N)		(regs_F.f[(N)])
#define SET_FPR_F(N,EXPR)	(regs_F.f[(N)] = (EXPR))
#define FPR_D(N)		(regs_F.d[(N) >> 1])
#define SET_FPR_D(N,EXPR)	(regs_F.d[(N) >> 1] = (EXPR))

/* miscellaneous register accessors */
#define SET_HI(EXPR)		(regs_HI = (EXPR))
#define HI			(regs_HI)
#define SET_LO(EXPR)		(regs_LO = (EXPR))
#define LO			(regs_LO)
#define FCC			(regs_FCC)
#define SET_FCC(EXPR)		(regs_FCC = (EXPR))

/* precise architected memory state accessor macros, all unsafe for speed  */
#define READ_WORD(SRC)							\
  (__MEM_READ_WORD(SRC))
#define READ_UNSIGNED_HALF(SRC)						\
  ((unsigned int)((unsigned short)__MEM_READ_HALF(SRC)))
#define READ_SIGNED_HALF(SRC)						\
  ((signed int)((signed short)__MEM_READ_HALF(SRC)))
#define READ_UNSIGNED_BYTE(SRC)						\
  ((unsigned int)((unsigned char)__MEM_READ_BYTE(SRC)))
#define READ_SIGNED_BYTE(SRC)						\
  ((unsigned int)((signed int)((signed char)__MEM_READ_BYTE(SRC))))

#define WRITE_WORD(SRC, DST)						\
  (__MEM_WRITE_WORD((DST), (unsigned int)(SRC)))
#define WRITE_HALF(SRC, DST)						\
  (__MEM_WRITE_HALF((DST), (unsigned short)((unsigned int)(SRC))))
#define WRITE_BYTE(SRC, DST)						\
  (__MEM_WRITE_BYTE((DST), (unsigned char)((unsigned int)(SRC))))

/* system call handler macro */
#define SYSCALL(INST)		(ss_syscall(mem_access, INST))

/* instantiate the helper functions in the '.def' file */
/* access global register */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)
#define CONNECT(OP)
#define IMPL
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
#undef IMPL



typedef	double				SimDouble;
typedef	float				SimSingle;
typedef	signed		int		SimWord;
typedef	signed		char	SimByte;
typedef	signed		int		SimRegister;
typedef	unsigned	int		SimAddress;

typedef	union	_SimInstruction
	{
	unsigned	int		ui[2];
	unsigned	short	us[4];
	unsigned	char	ub[8];
	}	SimInstruction;

typedef	long	long		Counter;

#	define	SimTextBase					0x00400000
#	define	SimDataBase					0x10000000
#	define	SimStackBsae				0x7fffc000

#	define	SimPageSize					4096
#	define	SimMaxUserEnvironment		16384

#	define	SimGlobalPointerRegister	28
#	define	SimStackPointerRegister		29
#	define	SimFramePointerRegister		30

#	define	SimInstructionOpCode(xxx)	(xxx->ui[0] & 0x000000ff)
#	define	SimInstructionPrePost(xxx)	((xxx->ui[0] & 0x0000ff00)>>8)
#	define		SimPrePostNone			0x00
#	define		SimPrePostPostInc		0x01
#	define		SimPrePostPostDec		0x02
#	define		SimPrePostPreInc		0x03
#	define		SimPrePostPreDec		0x04
#	define		SimPrePostPostDoubleInc	0x05
#	define		SimPrePostPostDoubleDec	0x06
#	define		SimPrePostPreDoubleInc	0x07
#	define		SimPrePostPreDoubleDec	0x08
#	define	SimInstructionRegisterSource(xxx)		((xxx->ui[1]>>24)&0xff)
#	define	SimInstructionRegisterTarget(xxx)		((xxx->ui[1]>>16)&0xff)
#	define	SimInstructionRegisterDestination(xxx)	((xxx->ui[1]>>8)&0xff)
#	define	SimInstructionShiftAmount(xxx)			(xxx->ui[1] & 0xff)
#	define	SimInstructionSignedImmediate(xxx)	\
				((int)((short)(xxx->ui[1] & 0xffff)))
#	define	SimInstructionUnsignedImmediate(xxx)		(xxx->ui[1] & 0xffff)
#	define	SimInstructionJumpTarget(xxx)			(xxx->ui[1]	& 0x3ffffff)
#	define	SimInstructionBreakCode(xxx)			(xxx->ui[1] & 0xfffff)

#	define	SimIntegerComputationFlag				0x00000001
#	define	SimFloatingPointComputationFlag			0x00000002
#	define	SimControlInstructionFlag				0x00000004
#	define	SimUnconditionalControlInstructionFlag	0x00000008
#	define	SimConditionalControlInstructionFlag	0x00000010
#	define	SimMemoryAccessFlag						0x00000020
#	define	SimMemoryLoadFlag						0x00000040
#	define	SimMemoryStoreFlag						0x00000080
#	define	SimDisplayedAddressingModeFlag			0x00000100
#	define	SimRegisterRegisterAddressingModeFlag	0x00000200
#	define	SimDirectAddressingModeFlag				0x00000400
#	define	SimTrapingInstructionFlag				0x00000800
#	define	SimLongLatencyInstructionFlag			0x00001000
#	define	SimDirectJumpFlag						0x00002000
#	define	SimIndirectJumpFlag						0x00004000
#	define	SimFunctionCallFlag						0x00008000
#	define	SimFloatingPointConditionalFlag			0x00010000


//#	define	MaxInstEntry	10000000
#	define	MaxInstEntry	0
// 3 being 100, 1000, and inf.
#	define	MaxH			(20+3)

Counter		iiCount=0,ffCount=0,llCount=0,ssCount=0,bbCount=0;
int			iCount=0, fCount=0, lCount=0, sCount=0, bCount=0;

Counter		IntegerH[MaxH][2];
Counter		FloatingH[MaxH][2];
Counter		LoadH[MaxH][2];
Counter		StoreH[MaxH][2];
Counter		BranchH[MaxH][2];

Counter	IntegerRegisters[255];
Counter	FloatingRegisters[255];

typedef	unsigned	int	bool;
#define	true	1
#define	false	0
bool	RegInstruction_Close();

Counter		LastBasicBlockInstruction = 0;

double	AvgBasicBlockSize, AvgBasicBlockSize2, TotalBasicBlocks;

//int	*IntegerRegisters	=	regs_R;

bool
Increment(
	Counter	Array[MaxH][2],
	int		ZeroOrOne,
	int		Value)
	{
	if(Value>=(MaxH-3))
		{
		if(Value<=100)
			++Array[MaxH-3][ZeroOrOne];
		else
		if(Value<=1000)
			++Array[MaxH-2][ZeroOrOne];
		else
			++Array[MaxH-2][ZeroOrOne];
		}
	else
		++Array[Value][ZeroOrOne];
	return(true);
	}

bool
RegInstruction_p(
	SimAddress	Address,
	int		OP,
	int		O1,
	int		O2,
	int		I1,
	int		I2,
	int		I3,
	int		FLAGS,
	Counter	InstructionCount)
	{
	if(!I1)	I1=200;
	if(!I2)	I2=200;
	if(!I3)	I3=200;

	if ( MaxInstEntry)
	if(	iCount>=MaxInstEntry||
		fCount>=MaxInstEntry||
		lCount>=MaxInstEntry||
		sCount>=MaxInstEntry||
		bCount>=MaxInstEntry	)
		{
		RegInstruction_Close();
		exit(1);
		}

	if(FLAGS & SimIntegerComputationFlag)
		{
		if(I1!=200)
			Increment(IntegerH, 0, InstructionCount - IntegerRegisters[I1]);
		else
			Increment(IntegerH, 0, 0);

		if(I2!=200)
			Increment(IntegerH, 1, InstructionCount - IntegerRegisters[I2]);
		else
			Increment(IntegerH, 1, 0);
		++iCount;
		}
	else
	if(FLAGS & SimFloatingPointComputationFlag)
		{
		if(I1!=200)
			Increment(FloatingH, 0, InstructionCount - FloatingRegisters[I1]);
		else
			Increment(FloatingH, 0, 0);

		if(I2!=200)
			Increment(FloatingH, 1, InstructionCount - FloatingRegisters[I2]);
		else
			Increment(FloatingH, 1, 0);
		++fCount;
		}
	else
	if(FLAGS & SimMemoryLoadFlag)
		{
		if(I2!=200)
			Increment(LoadH, 0, InstructionCount - IntegerRegisters[I2]);
		else
			Increment(LoadH, 0, 0);
		++lCount;
		}
	else
	if(FLAGS & SimMemoryStoreFlag)
		{
		if(I1!=200)
			Increment(StoreH, 0, InstructionCount - IntegerRegisters[I1]);
		else
			Increment(StoreH, 0, 0);
			
		if(I2!=200)
			Increment(StoreH, 1, InstructionCount - IntegerRegisters[I2]);
		else
			Increment(StoreH, 1, 0);
		++sCount;
		}
	else
	if(FLAGS & SimConditionalControlInstructionFlag)
		{	// branch flag
		if(I1!=200)
			Increment(BranchH, 0, InstructionCount - IntegerRegisters[I1]);
		else
			Increment(BranchH, 0, 0);
		if(I2!=200)
			Increment(BranchH, 1, InstructionCount - IntegerRegisters[I2]);
		else
			Increment(BranchH, 1, 0);
		++bCount;
		goto	BranchCalc;
		}
	else
		{
		double	Width;
BranchCalc:
		Width = InstructionCount - LastBasicBlockInstruction;
		AvgBasicBlockSize += Width;
		AvgBasicBlockSize2 += Width*Width;
		TotalBasicBlocks += 1.0;
		LastBasicBlockInstruction = InstructionCount;
		}
	return(true);
	}

bool
RegInstruction(
	SimAddress	Address,
	int		OP,
	int		O1,
	int		O2,
	int		I1,
	int		I2,
	int		I3,
	int		FLAGS,
	Counter	InstructionCount)
	{
	bool	Ret = RegInstruction_p(Address, OP, O1, O2, I1, I2, I3, FLAGS, InstructionCount);

	//printf("%d, %d <= %d, %d, %d\n", O1, O2, I1, I2, I3);
	IntegerRegisters[O1] = InstructionCount;
	IntegerRegisters[O2] = InstructionCount;
	FloatingRegisters[O1] = InstructionCount;
	FloatingRegisters[O2] = InstructionCount;
	if(FLAGS & SimIntegerComputationFlag)
		++iiCount;
	else
	if(FLAGS & SimFloatingPointComputationFlag)
		++ffCount;
	else
	if(FLAGS & SimMemoryLoadFlag)
		++llCount;
	else
	if(FLAGS & SimMemoryStoreFlag)
		++ssCount;
	else
		++bbCount;

	return(Ret);
	}

bool
PrintHistoGram(
	Counter		Array[MaxH][2],
	int			ZeroOrOne)
	{
	int	Index;
	double	Sum = 0.0, Sum2 = 0.0, TotalA = 0.0, v;
	Counter	Total = 0;
	for(Index=0;Index<MaxH;Index++)
		Total += Array[Index][ZeroOrOne];
	for(Index=0;Index<MaxH;Index++)
		printf("%.1f ", ((double)Array[Index][ZeroOrOne] / (double)Total * 100.0) );

	
	// determine avg and stddev for normal
	for(Index=1;Index<(MaxH-3);Index++)
		{
		Sum += Index*Array[Index][ZeroOrOne];
		Sum2 += Array[Index][ZeroOrOne]*Index*Index;
		TotalA += Array[Index][ZeroOrOne];
		}
	Sum = Sum / TotalA;
	Sum2 = Sum2 / (TotalA);
	v = sqrt(Sum2 - Sum*Sum);
	printf("Avg=%f,o=%f\n", (double)Sum, (double)v);	

	printf("\n");
	return(true);
	}

bool
RegInstruction_Close()
	{
	// calculate average for each
	int	Index = 0;
	double	Total;
	double	o;

	printf("I0: ");PrintHistoGram(IntegerH, 0);
	printf("I1: ");PrintHistoGram(IntegerH, 1);
	printf("F0: ");PrintHistoGram(FloatingH, 0);
	printf("F1: ");PrintHistoGram(FloatingH, 1);
	printf("L0: ");PrintHistoGram(LoadH, 0);
	printf("S0: ");PrintHistoGram(StoreH, 0);
	printf("S1: ");PrintHistoGram(StoreH, 1);
	printf("B0: ");PrintHistoGram(BranchH, 0);
	printf("B1: ");PrintHistoGram(BranchH, 1);

	Total = iiCount+ffCount+llCount+ssCount+bbCount;
	printf("Integer   : %lld (%f)\n", iiCount, ((double)iiCount/Total));
	printf("Floating  : %lld (%f)\n", ffCount, ((double)ffCount/Total));
	printf("Load      : %lld (%f)\n", llCount, ((double)llCount/Total));
	printf("Store     : %lld (%f)\n", ssCount, ((double)ssCount/Total));
	printf("Branch    : %lld (%f)\n", bbCount, ((double)bbCount/Total));

	AvgBasicBlockSize = AvgBasicBlockSize / TotalBasicBlocks;
	AvgBasicBlockSize2 = AvgBasicBlockSize2 / TotalBasicBlocks;
	o = sqrt(AvgBasicBlockSize2 - AvgBasicBlockSize*AvgBasicBlockSize);
	printf("Average Basic Block Size = %.2f, o=%.2f\n", AvgBasicBlockSize, o);
	return(true);
	}
/* start simulation, program loaded, processor precise state initialized */
void
sim_main(void)
{
	Counter	iCount;
#ifdef USE_JUMP_TABLE
  /* the jump table employs GNU GCC label extensions to construct an array
     of pointers to instruction implementation code, the simulator then uses
     the table to lookup the location of instruction's implementing code, a
     GNU GCC `goto' extension is then used to jump to the inst's implementing
     code through the op_jump table; as a result, there is no need for
     a main simulator loop, which eliminates one branch from the simulator
     interpreter - crazy, no!?!? */

  /* instruction jump table, this code is GNU GCC specific */
  static void *op_jump[/* max opcodes */256] = {
    &&opcode_NA, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
    &&opcode_##OP,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
    &&opcode_##OP,
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
  };
#endif /* USE_JUMP_TABLE */

  /* register allocate PC and next PC */
  register SS_ADDR_TYPE local_regs_PC = regs_PC, local_next_PC;
  /* register allocate integer register file base pointer address */
  register SS_WORD_TYPE *local_regs_R = regs_R;
  /* register allocate level one memory page map */
#undef mem_table
  register char **local_mem_table = mem_table;
#define mem_table local_mem_table
#ifdef USE_JUMP_TABLE
  /* register allocate local jump table base pointer address */
  register void **local_op_jump = op_jump;
#endif /* USE_JUMP_TABLE */
  /* register allocate instruction buffer */
  register SS_INST_TYPE inst;

  fprintf(stderr, "sim: ** starting *fast* functional simulation **\n");

  /* must have natural byte/word ordering */
  if (sim_swap_bytes || sim_swap_words)
    fatal("sim: *fast* functional simulation cannot swap bytes or words");

#ifdef USE_JUMP_TABLE

  /* set up initial default next PC */
  local_next_PC = regs_PC;

  /* load instruction */
  inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, local_next_PC);

  /* jump to instruction implementation */
  goto *local_op_jump[SS_OPCODE(inst)];

#ifndef NO_INSN_COUNT
#define INC_INSN_CTR()	sim_num_insn++
#else /* !NO_INSN_COUNT */
#define INC_INSN_CTR()	/* nada */
#endif /* NO_INSN_COUNT */

#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
  opcode_##OP:								\
    /* maintain $r0 semantics */					\
    local_regs_R[0] = 0;						\
    /* keep an instruction count */					\
    INC_INSN_CTR();							\
    /* execute next instruction */					\
    local_regs_PC = local_next_PC;					\
    /* set up default next PC */					\
    local_next_PC += 8;							\
    /* execute the instruction */					\
    EXPR;								\
    /* get the next instruction */					\
    inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, local_next_PC);		\
    /* jump to instruction implementation */				\
    goto *local_op_jump[SS_OPCODE(inst)];
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
  opcode_##OP:								\
    panic("attempted to execute a linking opcode");
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
 opcode_NA:
    panic("attempted to execute a bogus opcode");

  /* should not get here... */
  panic("exited sim-fast main loop");

#else /* !USE_JUMP_TABLE */

  /* set up initial default next PC */
  local_next_PC = regs_PC;
	iCount = 0;
  while (TRUE)
    {
	++iCount;
      /* maintain $r0 semantics */
      local_regs_R[0] = 0;

      /* keep an instruction count */
#ifndef NO_INSN_COUNT
      sim_num_insn++;
#endif /* !NO_INSN_COUNT */

      inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, local_regs_PC);

		/*RegInstruction(regs_PC,OP,O1,O2,I1,I2,I3,FLAGS, iCount);*/
	if(SS_OPCODE(inst)==SYSCALL)
		if(regs_R[2] == SS_SYS_exit)
			RegInstruction_Close();

      switch (SS_OPCODE(inst)) {
#define DNA 200
#define DFCC    201
#define DGPR(xxx)   (xxx)
#define DCGPR(xxx)  (xxx)
#define DGPR_D(xxx) (xxx)
#define DFPR(xxx)   ((xxx)+32)
#define DFPR_L(xxx) ((xxx)+32)
#define DFPR_D(xxx) ((xxx)+32)
#define DFPR_F(xxx) ((xxx)+32)
#define DHI 202
#define DLO 203

#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
      case OP:		\
		RegInstruction(regs_PC,OP,O1,O2,I1,I2,I3,FLAGS, iCount);\
        EXPR;		\
        break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
      case OP:											\
        panic("attempted to execute a linking opcode");
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
      }

      /* execute next instruction */
      local_regs_PC = local_next_PC;
      local_next_PC += SS_INST_SIZE;
    }

#endif /* USE_JUMP_TABLE */

}
